unit Fliplay;

{
  This Delphi 1.0 component is designend to display FLI files
  Copyright (C) 1996  Paul Toth

  TothPaul@mygale.org
  http://www.mygale.org/~tothpaul

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

}

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, ExtCtrls, DsgnIntf;

type
 { Replace the TBitmapInfo for dib_pal_colors on StretchDiBits API }
 TPalBitmapInfo=record
  bmiHeader:TBitmapInfoHeader;
  bmiColors:Array[0..255] of Word;
 end;

  { FLI Header }
  TFliHeader = Record
   Padding1 : LongInt;
   ID       : Word;
   Frames   : Word;
   Padding2 : LongInt;
   Padding3 : LongInt;
   Speed    : Word;
   Padding4 : ARRAY[1..110] OF Char; {Pad to 128 Bytes}
  end;
  { Frames Header }
  TFrameHeader = Record
    Size     : LongInt;
    Padding1 : Word;
    Chunks   : Word;
    Padding2 : Array[1..8] of Char; {Pad to 16 Bytes}
  end;
  { Chunk }
  TChunk = record
   Size:word;
   Zero:word; { ??? }
   Kind:word;
   Data:byte;
  end;
  { Color entry }
  RGB = record
   Red,Green,Blue:byte;
  end;
  { Palette entries }
  TPalettePacket=record
   FirstColor:byte;
   ColorCount:byte;
   Colors:Array[0..255] of RGB;
  end;
  { Fli Palette entries }
  TFliPalette=record
   PacketCount:word;
   FirstPacket:TPalettePacket;
  end;

  TDataPacket = Array[0..256] of byte;
  { Draw Packet }
  TDrawPacket=record
   IncX:byte;
   Size:byte;
   Data:TDataPacket;
  end;
  { TDrawLines }
  TDrawLines = record
   PacketCount:byte;
   FirstPacket:TDrawPacket;
  end;
  { Fli Draw }
  TFliDraw=record
   DrawLine:word;
   DrawHeight:word;
   FirstLine:TDrawLines;
  end;
  { Brun Packet }
  TBrunPacket=record
   Size:byte;
   Data:TDataPacket;
  end;
  { Fli Brun  }
  TFliBrun=record
   PacketCount:byte;
   FirstPacket:TDrawPacket;
  end;

  TBitmapData = Array [0..199,0..319] of byte; { Virtual 320x200 screen }

  TFliPlayer = class(TGraphicControl)
  private
    { Private-dclarations }
    fActive:boolean;   { FLI Running ? }
    fFliFile:string;   { FLI File Name }
    fOpen:boolean;     { File open }
    fLoading:boolean;  { Switch on while loading a frame }
    fInterval:word;    { Timer duration }
    fLoop:boolean;     { loop at last frame }
    fImageX:integer;
    fImageY:integer;
    fImageH:integer;
    fImageW:integer;

    PalCount:Word;    { Number of Palette Chunck : skip PaletteEntries if possible }
    FirstPal:Boolean; { Flag for Palette Chunck count }

    Fli:File;           { File handle  }
    Frame:Word;         { current frame }
    FrameCount:Word;    { total frames }
    Speed:Word;         { speed of FLI }
    FirstFrame:LongInt; { FilePos(FirstFrame) }
    BitmapData:^TBitmapData;  { Drawing Frame }
    Handle:HWnd;        { Timer Handle }
    fPalette:HPalette;   { 256 colors palette }
    fOldPalette:HPalette;
    BitsInfo:^TPalBitmapInfo; { Bitmap information to copy Virtual Screen on to Device }

    Procedure SetActive(Value:boolean);
    Procedure SetFliFile(Value:string);
    Procedure SetInterval(Value:word);
    Procedure SetImageX(Value:integer);
    Procedure SetImageY(Value:integer);
    Procedure SetImageH(Value:integer);
    Procedure SetImageW(Value:integer);
    Procedure WndProc(Var Msg:TMessage); { Timer callback }

  protected
    { Protected-dclarations }
    Constructor Create(AOwner:TComponent); override;
    Destructor Destroy; override;
    Procedure CreatePalette256;
    Procedure NextFrame;
    Procedure DrawFrame(Var FrameData; Chunks:Word);
    Procedure FliPalette(Var PaletteData);
    Procedure FliDraw(Var DrawData);
    Procedure FliBrun(Var BrunData);
    Procedure FliCopy(Var CopyData);
    Function UnPackData(Var x:integer; y:integer; Var Data:TDataPacket;Size:Word):word;
    Procedure Paint; override;
  public
    { Public-dclarations }
  published
    { Published-dclarations }
    Property Active:boolean read fActive write SetActive;
    Property FileName:string read fFliFile write SetFliFile;
    Property Interval:word read fInterval write SetInterval stored false;
    Property Loop:boolean read fLoop write fLoop;
    Property ImageX:integer read fImageX write SetImageX;
    Property ImageY:integer read fImageY write SetImageY;
    Property ImageH:integer read fImageH write SetImageH;
    Property ImageW:integer read fImageW write SetImageW;
    { Inherited properties }
    Property Align;
  end;

  TFileNameProperty = Class (TStringProperty)
   Function GetAttributes:TPropertyAttributes; override;
   Procedure Edit; override;
  end;

procedure Register;

implementation

Constructor TFliPlayer.Create;
 begin
  inherited Create(AOwner);
  { Default size }
  Width:=320;
  Height:=200;

  CreatePalette256;

  { Bitmap }
  New(BitmapData); { virtual screen }
  fImageX:=1;
  fImageY:=1;
  fImageW:=320;
  fImageH:=200;

  { New bitmap description }
  getmem(bitsinfo,sizeof(tbitmapinfoheader)+256*SizeOf(Word));
  with bitsinfo^.bmiheader do begin
   bisize:=40;
   biwidth:=320;
   biheight:=200;
   biplanes:=1;
   bibitcount:=8;
   bicompression:=BI_RGB;
   bisizeimage:=0;
   bixpelspermeter:=0;
   biypelspermeter:=0;
   biclrused:=256;
   biclrimportant:=0;
  end;

  { Timer Handle }
  Speed    := 0;
  Handle   := AllocateHWnd(WndProc);
  fOpen    := False;
  fLoading := False;
 end;

Procedure TFliPlayer.CreatePalette256;
 var
  Pal256:pLogPalette;
 begin
  { 256 color Palette }
  Try
   GetMem(Pal256,4+256*4);
   Pal256^.palVersion:=$300;
   Pal256^.palNumEntries:=256;
   FillChar(Pal256^.palPalEntry,4*256,0);
   fPalette:=CreatePalette(Pal256^);
  Finally
   FreeMem(Pal256,4+256*4);
  end;
 end;

Destructor TFliPlayer.Destroy;
 begin
  DeallocateHWnd(Handle); { remove timer }
  Active:=False;
  if fOpen then CloseFile(Fli);
  DeleteObject(fPalette);
  freemem(bitsinfo,sizeof(tbitmapinfoheader)+256*SizeOf(Word));
  Dispose(BitmapData);
  inherited Destroy;
 end;

Procedure TFliPlayer.SetActive;
 begin
  if fActive=Value then exit;
  fActive:=Value;
  if (Value)and(FileName<>'') then Interval:=Speed else Interval:=0;
 end;

Procedure TFliPlayer.SetFliFile;
 Var
  Header:TFliHeader;
 begin
  if fFliFile=Value then exit;
  Active:=False;
  if fOpen then begin CloseFile(Fli); fOpen:=False; end;
  fFliFile:=Value;
  if value='' then exit;
  Try
   FileMode:=0;
   AssignFile(Fli,Value); Reset(Fli,1);
   fopen:=True;
   BlockRead(Fli,Header,SizeOf(Header));
   If Header.ID=$AF11 then begin
    Speed:=Header.Speed;
    FrameCount:=Header.Frames;
    Frame:=0;
    PalCount:=0;
    FirstPal:=True;
     FirstFrame:=FilePos(Fli);
     NextFrame;
    FirstPal:=False;
   end;
  Except
   On E:Exception do begin
    Active:=False;
    ShowMessage(E.Message);
   End;
  end;
 end;

Procedure TFliPlayer.SetInterval;
 begin
  If fInterval=Value then exit;
  if fInterval>0 then KillTimer(Handle,1);
  fInterval:=Value;
  if fInterval>0 then SetTimer(Handle,1,fInterval,nil);
 end;

Procedure TFliPlayer.SetImageX;
 begin
  if fImageX=Value then exit;
  if Value>320 then exit;
  fImageX:=Value;
  if Value+fImageW>320 then fImageW:=320-fImageX;
  Paint;
 end;

Procedure TFliPlayer.SetImageY;
 begin
  if fImageY=Value then exit;
  if Value>200 then exit;
  fImageY:=Value;
  if Value+fImageH>200 then fImageH:=200-fImageY;
  Paint;
 end;

Procedure TFliPlayer.SetImageH;
 begin
  if fImageH=Value then exit;
  if Value+fImageY>200 then fImageH:=200-fImageY else fImageH:=Value;
  Paint;
 end;

Procedure TFliPlayer.SetImageW;
 begin
  if fImageW=Value then exit;
  if Value+fImageX>320 then fImageW:=320-fImageX else fImageW:=Value;
  Paint;
 end;

Procedure TFliPlayer.WndProc;
 begin
  with Msg do
    if Msg = WM_TIMER then
      try
        if not fLoading then begin
         fLoading:=True;
         NextFrame;
         fLoading:=False;
        end else MessageBeep(0);
      except
        Application.HandleException(Self);
      end
    else
      Result := DefWindowProc(Handle, Msg, wParam, lParam);
 end;

Procedure TFliPlayer.NextFrame;
 Var
  FrameHeader:TFrameHeader;
  FrameSize:Word;
  FrameData:pointer;
 begin
  Try
   BlockRead(Fli,FrameHeader,SizeOf(FrameHeader));
   FrameSize := FrameHeader.Size - SizeOf(FrameHeader);
   GetMem(FrameData,FrameSize);
    BlockRead(Fli,FrameData^,FrameSize);
    DrawFrame(FrameData^,FrameHeader.Chunks);
   FreeMem(FrameData,FrameSize);

   Paint;

   Inc(Frame);
   if (Frame=FrameCount) then begin
    Seek(Fli,FirstFrame);
    Frame:=1;
    Active:=Loop;
   end;
  except
   on e:Exception do begin Active:=False; Showmessage(E.Message); end;
  end;
 end;

Procedure TFliPlayer.DrawFrame;
 Var
  Chunk:^TChunk;
 begin
  Chunk:=@FrameData;
  While Chunks>0 do begin
   With Chunk^ do begin
    Case Kind of
     $0B : If (FirstPal)or(PalCount>1) then FliPalette(Data) { else Colors allready set };
     $0C : FliDraw(Data);
     $0D : FillChar(BitmapData^,SizeOf(TBitmapData),0);
     $0F : FliBrun(Data);
     $10 : FliCopy(Data);
    else begin
     showmessage('Invalide Chunk type $'+inttohex(Kind,2)+' ('+IntToStr(Chunk^.Size)+' octets)');
     active:=false;
    end;
    end;
    If Zero<>0 then begin
     showmessage('Chunk Size trop grand : '+IntToStr(Zero));
     active:=false;
    end;
    Chunk:=Ptr(Seg(Chunk^),Ofs(Chunk^)+Size);
    dec(Chunks);
   end;
  end;
 end;

type
 tColorIndex=Array[0..255] of word;

Procedure TFliPlayer.FliPalette;
 Var
  Color:word;
  Packets:integer;
  Packet:^TPalettePacket;
  Count:integer;
  i:integer;
  Colors:Array[0..255] of TPaletteEntry;
 begin
  If FirstPal then Inc(PalCount);
  With TFliPalette(PaletteData) do begin
   Color:=0;
   Packets:=PacketCount;
   Packet:=@FirstPacket;
   While Packets>0 do begin dec(Packets);
    Inc(Color,Packet^.FirstColor);
    Count:=Packet^.ColorCount;
    if Count=0 then Count:=256;
    i:=0;
    while i<count do begin
     BitsInfo^.bmiColors[color+i]:=Color+i;
     Colors[color+i].peRed  :=Packet^.Colors[i].red shl 2;
     Colors[color+i].peGreen:=Packet^.Colors[i].Green shl 2;
     Colors[color+i].peBlue :=Packet^.Colors[i].Blue shl 2;
     Colors[color+i].peFlags:=PC_NOCOLLAPSE;
     inc(i);
    end;
    SetPaletteEntries(fPalette,Color,Count,Colors[Color]);
   end;
   Packet:=@Packet^.Colors[i];
  end;
 end;

Function TFliPlayer.UnPackData;
 begin
  if Size>127 then begin { rptition }
   FillChar(BitmapData^[y,x],256-Size,Data[0]);
   Inc(x,256-Size);
   Result:=1;
  end else begin { srie de points }
   Move(Data,BitmapData^[y,x],Size);
   inc(x,Size);
   Result:=Size;
  end;
 end;

Procedure TFliPlayer.FliDraw;
 Var
  Packet:^TDrawPacket;
  x,y:integer;
  h,packets:word;
  DrawLines:^TDrawLines;
 begin
  With TFliDraw(DrawData) do begin
   y:=199-DrawLine;
   h:=DrawHeight;
   DrawLines:=@FirstLine;
  end;
  while h>0 do
   With DrawLines^ do begin
    Packets:=PacketCount; { nombre de packets }
    Packet:=@FirstPacket;
    x:=0;
    while Packets>0 do
     with Packet^ do begin
      Inc(x,IncX);
      Packet:=@Data[UnPackData(x,y,Data,Size)];
      Dec(Packets);
     end;
    dec(y); { ligne suivante }
    DrawLines:=@Packet^;
    dec(h);
   end;
 end;

Procedure TFliPlayer.FliBrun;
 Var
  x,y:integer;
  Packets:integer;
  Packet:^TBrunPacket;
  FliBrun:^TFliBrun;
 begin
  FliBrun:=@BrunData;
  y:=199;
  while y>=0 do
   With FliBrun^ do begin
    Packets:=PacketCount;
    Packet:=@FirstPacket;
    x:=0;    { dbut de ligne }
    while Packets>0 do
     With Packet^ do begin
      Packet:=@Data[UnPackData(x,y,Data,256-Size)];
      dec(Packets);
     end;
    dec(y); { ligne suivante }
    FliBrun:=@Packet^;
   end;
 end;

Procedure TFliPlayer.FliCopy;
 Var
  i:integer;
 begin
  i:=199;
  while i>=0 do begin
   Move(TBitmapData(CopyData)[i,0],BitmapData^[199-i,0],320);
   dec(i);
  end;
 end;

Procedure TFliPlayer.Paint;
 Var
  BitmapInfo:^TBitmapInfo;
 begin
  SetStretchBltMode(Canvas.Handle, STRETCH_DELETESCANS);
  SelectPalette(Canvas.Handle,fPalette,False);
  RealizePalette(Canvas.Handle);
  BitmapInfo:=pointer(BitsInfo);
  StretchDIBits(Canvas.Handle,0,0,Width,Height,
                fImageX,200-(fImageH+fImageY),fImageW,fImageH,
                BitmapData,BitmapInfo^,dib_pal_colors,Canvas.CopyMode);
 end;

{ FileName Property }
Function TFileNameProperty.GetAttributes;
 begin
  Result:=[paDialog];
 end;

Procedure TFileNameProperty.Edit;
 Var
  OpenDlg:TOpenDialog;
 begin
  try
   OpenDlg:=TOpenDialog.Create(Application);
   OpenDlg.Options:=[ofHideReadOnly];
   OpenDlg.DefaultExt:='FLI';
   OpenDlg.Filter := 'Animation FLI|*.FLI';
   if OpenDlg.Execute then SetValue(OpenDlg.FileName);
  finally
   OpenDlg.Free;
  end;
 end;

procedure Register;
begin
 RegisterComponents('Exemples', [TFliPlayer]);
 RegisterPropertyEditor(TypeInfo(String),TFliPlayer,'FileName',TFileNameProperty);
end;

end.
